# Package: PENTAtrainer2
# Version: 1.0
# Filename: annotate.praat
#
# Author: Santitham Prom-on
# Last updated: 10 July 2015
#
# Objective: annotate input speech files iteratively. This process includes 
# pulse marking, pitch extraction, boundary identification, functional marking

## Macro ##
taskEdit = 1
taskResume = 2
taskCreate = 3
taskAdd = 4
taskRemove = 5
#

##### Input Module - Begin #####

#Select the working directory
directory$ = chooseDirectory$ ("Choose a directory to annotate")
if directory$ = ""
	#[EXIT_00] Choose to not Continue
	exit
endif

#Check whether the selected directory contain .wav or .WAV files
Create Strings as file list... FileList 'directory$'/*.wav
numberOfFiles = Get number of strings
if !numberOfFiles
	select Strings FileList
	Remove
	Create Strings as file list... FileList 'directory$'/*.WAV
	numberOfFiles = Get number of strings
endif

#[Error checking] No files in the directory	
if numberOfFiles = 0
	#[EXIT_A01] Sound files not found
	exit Error Code A01: No sound files in the directory
else
	Write to raw text file... 'directory$'/FileList.txt
endif

Create Table with column names... f0check 'numberOfFiles' Name Flag

#Read annotation files
Create Strings as file list... annot_list 'directory$'/*.annotation_short
n_annot = Get number of strings
Remove

# Main option window
beginPause("Annotation - Options")
	comment("Select the annotation task")
	optionMenu("Task:", 1)
	if numberOfFiles = n_annot
		#Fully annotated
		option("Edit annotation")
		option("Create new annotation")
		option("Add new function")
		option("Remove function")
		boolean("Process all files without pause",0)
		boolean("Edit F0 analysis options",0)
	elsif n_annot > 0 
		#Partially annotated
		option("Resume annotation")
		option("Create new annotation")
		process_all_files_without_pause = 0
		edit_F0_analysis_options = 0
	else	
		#Not annotated
		option("Create new  annotation")
		boolean("Edit F0 analysis options",0)
		process_all_files_without_pause = 0
	endif
	boolean("Inspect pulse",0)
clicked = endPause("Cancel","Start",2,1)

#[EXIT_00] Choose to not Continue
if clicked = 1
	exit
endif

#Reassign task variable based on the annotation status
if numberOfFiles = n_annot
	#Fully annotated
	if task = 1
		task = taskEdit
	elsif task = 2
		task = taskCreate
	elsif task = 3
		task = taskAdd
	else
		task = taskRemove
	endif
elsif n_annot > 0 
	#Partially annotated
	if task = 1
		task = taskResume
	else
		task = taskCreate
	endif
else	
	#Not annotated
	task = taskCreate
endif

# If users want to edit the F0 analysis option
if fileReadable("'directory$'/config.txt")
	##This directory has been annotated before
	
	Read from file... 'directory$'/config.txt
	c_nrows = Get number of rows
	c_ncols = Get number of columns
	
	#[EXIT_C01] Wrong config file format
	if c_nrows <> 7 and c_ncols <> 2
		exit Error Code C01: Wrong config file format
	endif
	
	#Read values from the existing config file
	c_np = Get value... 1 value
	c_minf0 = Get value... 2 value
	c_maxf0 = Get value... 3 value
	c_fsr = Get value... 4 value
	c_pl = Get value... 5 value
	c_fo = Get value... 6 value
	c_sit = Get value... 7 value
	Remove
	
	#Create a form with values read from the existing config file
	if edit_F0_analysis_options
		beginPause ("Edit F0 Analysis Options")
			integer("left F0 range (Hz)",'c_minf0')
			integer("right F0 range (Hz)", 'c_maxf0')
			integer("Number of point per interval",'c_np')
			integer("F0 sample rate (Hz)",'c_fsr')
			integer("Perturbation length (ms)",'c_pl')
			real("Final offset",'c_fo')
			boolean("Set initial time for normf0 to 0",'c_sit')
		clicked = endPause("Cancel","Continue",2,1)
		
		#[EXIT_00] Choose to not continue
		if clicked = 1
			exit
		endif
		
		npoints = number_of_point_per_interval
		minf0 = left_F0_range
		maxf0 = right_F0_range
	else
		#or pass the values to the working variables
		npoints = c_np
		minf0 = c_minf0
		maxf0 = c_maxf0
		f0_sample_rate = c_fsr
		perturbation_length = c_pl
		final_offset = c_fo
		set_initial_time_for_normf0_to_0 = c_sit
	endif
else
	##This directory has not been annotated before, create
	
	if edit_F0_analysis_options
		#Prompt the window to display analysis options
		beginPause ("Edit F0 Analysis Options")
			integer("left F0 range (Hz)",30)
			integer("right F0 range (Hz)", 400)
			integer("Number of point per interval",10)
			integer("F0 sample rate (Hz)",100)
			integer("Perturbation length (ms)",0)
			real("Final offset",0)
			boolean("Set initial time for normf0 to 0",1)
		clicked = endPause("Cancel","Continue",2,1)
		
		#[EXIT_00] Choose to not continue
		if clicked = 1
			exit
		endif
		
		npoints = number_of_point_per_interval
		minf0 = left_F0_range
		maxf0 = right_F0_range
	else
		#Edit F0 analysis option is not selected. Set all options to default values
		npoints = 10
		minf0 = 30
		maxf0 = 400
		f0_sample_rate = 100
		perturbation_length = 0
		final_offset = 0
		set_initial_time_for_normf0_to_0 = 1
	endif
endif

#Create and save the Table object to be saved with the input values
Create Table with column names... config 7 parameter value
Set string value... 1 parameter npoints
Set numeric value... 1 value 'npoints'

Set string value... 2 parameter minf0
Set numeric value... 2 value 'minf0'

Set string value... 3 parameter maxf0
Set numeric value... 3 value 'maxf0'

Set string value... 4 parameter f0_sample_rate
Set numeric value... 4 value 'f0_sample_rate'

Set string value... 5 parameter perturbation_length
Set numeric value... 5 value 'perturbation_length'

Set string value... 6 parameter final_offset
Set numeric value... 6 value 'final_offset'

Set string value... 7 parameter set_initial_time_for_normf0_to_0
Set numeric value... 7 value 'set_initial_time_for_normf0_to_0'

Save as tab-separated file... 'directory$'/config.txt
Remove

##### Input Module - End #####

##### Task Module - Begin #####
if task = taskEdit
	# Scenario: The annotation is completed. Users want to edit the existing annotation.

	#Select the starting file to annotate
	beginPause("Edit existing annotation")
		integer("Input File No (from 'n_annot')", 1)
	clicked = endPause("Cancel","Continue",2,1)
	
	#[EXIT_00] Choose to not continue
	if clicked = 1
		exit
	endif
	
	#[EXIT_A02] Error: Input File No must be less than or equal to 'n_annot'
	if input_File_No > n_annot
		exit Error Code A02: Input File No must be less than or equal to 'n_annot'
	endif
	
;	if not fileReadable("'directory$'/function_profile.txt")
		#Read from the first file in FileList.txt to fine name$
		select Strings FileList
		fileName$ = Get string... 1
		name$ = fileName$ - ".wav" - ".WAV"
		
		#Read from the name$.annotation_short to find the list of communicative functions
		Read from file... 'directory$'/'name$'.annotation_short
		n_func = Get number of tiers
		list_func$ = ""		
		func_name$ = Get tier name... 1
		list_func$ = func_name$
		if n_func > 1
			for i from 2 to n_func
				func_name$ = Get tier name... i
				list_func$ = list_func$ + " " + func_name$
			endfor
		endif
		Remove
		
		#Create a table containing the functional profile
		#   (1) list of communicative function
		#   [depreciated] (2) the file extension to import
		#   [depreciated] (3) the tier number to import
		#   [depreciation note: (2) and (3) are replaced by import.praat
		Create Table with column names... functionProfile 3 parameter value
		Set string value... 1 parameter list_func
		Set string value... 1 value 'list_func$'
		Set string value... 2 parameter copy_from
		Set string value... 2 value n/a
		Set string value... 3 parameter copy_tier
		Set numeric value... 3 value 0
		Save as tab-separated file... 'directory$'/function_profile.txt
		Remove
;	endif
	
elsif task = taskResume
	# Scenario: The annotation is not finished. Users want to resume the ongoing annotation

	beginPause("Do you want to resume current annotation?")
	opt = endPause("No","Yes",2,1);
	if opt == 2
		#Users choose to resume
		input_File_No = n_annot + 1
		if fileReadable("'directory$'/function_profile.txt")
			Read from file... 'directory$'/function_profile.txt
			c_nrows = Get number of rows
			c_ncols = Get number of columns
			#[EXIT_C02] Wrong function profile file format
			if c_nrows <> 3 and c_ncols <> 2
				exit Error Code C02: Wrong function profile file format
			endif
			list_func$ = Get value... 1 value
			copy_from$ = Get value... 2 value
			copy_tier = Get value... 3 value
			task = taskCreate
			Remove
		else
			#[EXIT C03] Cannot find function profile
			exit Error Code C03: Cannot find function profile
		endif
	else
		#Users do not choose to resume
		exit
	endif

elsif task = taskCreate
	# Scenario: Users want to create new annotation
	input_File_No = 1
	beginPause("Create new annotation")
		sentence("List of communicative functions", "Tone Focus Sentence")
#		sentence("Copy boundary from","")
#		integer("Boundary tier",1);
#		boolean("Copy tier content",0)
	clicked = endPause("Cancel","Continue",2,1)
	
	#[EXIT_00] Choose to not continue
	if clicked = 1
		exit
	endif
	
	list_func$ = list_of_communicative_functions$
	copy_from$ = "" ;copy_boundary_from$
	copy_tier = 1 ;boundary_tier
	copy_content = 1 ;copy_tier_content
	
	### Error Checking
		#[EXIT_A03] No input communicative functions
		if list_func$ == ""
			exit Error Code A03: No input communicative functions
		endif
		
		if copy_from$ <> ""
			if not startsWith(copy_from$,".")
				copy_from$ = "." + copy_from$
			endif
			Create Strings as file list... b_list 'directory$'/*'copy_from$'
			n_boundary = Get number of strings
			
			#[EXIT_A04] Cannot find source boundary files to copy from
			if n_boundary = 0
				exit Error Code A04: Cannot find source boundary files to copy from
			endif
			
			#[EXIT_A05] Number of source boundary files must equal number of sound files
			if n_boundary <> numberOfFiles
				exit Error Code A05: Number of source boundary files must equal number of sound files
			endif
			
			select Strings b_list
			fileName$ = Get string... 1
			Read from file... 'directory$'/'fileName$'
			fullName$ = selected$()
			type$ = extractWord$ (fullName$, "")
			
			#[EXIT_A06] Wrong boundary file type (TextGrid is required)
			if type$ <> "TextGrid"
				exit Error Code A06: Wrong boundary file type. TextGrid is required instead 'type$' was found.
			else
				nTierBoundaryFile = Get number of tiers
				#[EXIT_A07] Selected tier to copy from exceed the number of existing tiers
				if copy_tier > nTierBoundaryFile
					exit Error Code A07: Selected tier to copy from exceed the number of existing tiers
				endif
				
				#[EXIT_A08] Selected tier to copy from must be more than zero
				if copy_tier < 1
					exit Error Code A08: Selected tier to copy from must be more than zero
				endif
			endif
			
						
			plus Strings b_list
			Remove
		endif
	### End of Error Checking
	
	Create Table with column names... functionProfile 3 parameter value
	Set string value... 1 parameter list_func
	Set string value... 1 value 'list_func$'
	Set string value... 2 parameter copy_from
	Set string value... 2 value 'copy_from$'
	Set string value... 3 parameter copy_tier
	Set numeric value... 3 value 'copy_tier'
	Save as tab-separated file... 'directory$'/function_profile.txt
	Remove
	
	#Ask if users want to replace all annotations
	if n_annot > 0
		beginPause("Annotation exist. Replace all?")
		replace_option = endPause("No","Yes",1)
		if replace_option = 1
			exit
		else
			#Delete existing file before begin annotation
			Create Strings as file list... annot_list 'directory$'/*.annotation_short
			n_annot = Get number of strings
			for i from 1 to n_annot
				select Strings annot_list
				fileName$ = Get string... i
				name$ = fileName$ - ".annotation_short"
				call deleteAuxFiles 'name$'
			endfor
			n_annot = 0
			Remove
		endif
	endif

elsif task == taskAdd
	# Scenario: Annotation is finished. Users want to add additional layer into the annotation
	input_File_No = 1
	beginPause("Add new function")
	sentence("Function to be added","")
	endPause("Continue",1)
	function_name$ = function_to_be_added$
	if function_name$ == ""
		#[EXIT_A10] Input function cannot be empty
		exit Error Code A10: Input function cannot be empty
	else
		if fileReadable("'directory$'/function_profile.txt")
			Read from file... 'directory$'/function_profile.txt
			c_nrows = Get number of rows
			c_ncols = Get number of columns
			#[EXIT_C02] Wrong function profile file format
			if c_nrows <> 3 and c_ncols <> 2
				exit Error Code C02: Wrong function profile file format
			endif
			list_func$ = Get value... 1 value
			list_func$ = list_func$ + " " + function_name$
			Set string value... 1 value 'list_func$'
			Save as tab-separated file... 'directory$'/function_profile.txt
			Remove
		else
			#[EXIT C03] Cannot find function profile
			exit Error Code C03: Cannot find function profile
		endif
		
#		if fileReadable("'directory$'/function.txt")
#			Read Table from tab-separated file... 'directory$'/function.txt
#			c_nrows = Get number of rows
#			c_ncols = Get number of columns
#			#[EXIT_C05] Wrong function list file format
#			if c_ncols <> 1
#				exit Error Code C05: Wrong function list file format
#			endif
#			Append row
#			new_row = c_nrows+1
#			printline 'function_name$'
#			Set string value... new_row FUNCTION 'function_name$'
#			Save as tab-separated file... 'directory$'/function.txt
#			Remove
#		else
#			#[EXIT C04] Cannot find function list
#			exit Error Code C03: Cannot find function list
#		endif
	endif

else task == taskRemove
	# Scenario: Annotation is completed. Users want to remove one layer out

	# Prompt the Remove option window to users
	input_File_No = 1
	beginPause("Remove existing function")
	integer("Tier number",0)
	endPause("Continue",1)
	
	if fileReadable("'directory$'/function_profile.txt")
		Read from file... 'directory$'/function_profile.txt
		c_nrows = Get number of rows
		c_ncols = Get number of columns
		#[EXIT_C02] Wrong function profile file format
		if c_nrows <> 3 and c_ncols <> 2
			exit Error Code C02: Wrong function profile file format
		endif
		list_func$ = Get value... 1 value
		Create TextGrid... 0 1 "'list_func$'" 
		Rename... newListFunc
		Remove tier... tier_number
		new_list_func$ = ""
		new_nrows = Get number of tiers
		
		if new_nrows > 0
			func_name$ = Get tier name... 1
			new_list_func$ = func_name$
			if new_nrows > 1
				for i from 2 to new_nrows
					func_name$ = Get tier name... i
					new_list_func$ = new_list_func$ + " " + func_name$
				endfor
			endif
		else
			# EXIT_A11: Cannot remove function tier
			exit Error Code A11: Cannot remove function tier
		endif
		select TextGrid newListFunc
		Remove
		select Table function_profile
		Set string value... 1 value 'new_list_func$'
		Save as tab-separated file... 'directory$'/function_profile.txt
		Remove
	else
		#[EXIT C03] Cannot find function profile
		exit Error Code C03: Cannot find function profile
	endif
endif	
##### Task Module - End #####

# for each file (from input_File_No to numberOfFiles), call Labeling 
for current_file from input_File_No to numberOfFiles
	select Strings FileList
	fileName$ = Get string... current_file
	name$ = fileName$ - ".wav" - ".WAV"
	call Labeling 'fileName$'
endfor

call ReportError

if cflag = 0
	printline === Error has occurred: Annotation incomplete ===
else
	printline === Annotation completed ===
endif

select Strings FileList
plus Table f0check
Remove
#===End of Main===
#####

##########Other Procedures##########

procedure Labeling file_name$
	#Prepare the files necessary to annotation
	Read from file... 'directory$'/'file_name$'
	if task = taskEdit
		if fileReadable ("'directory$'/'name$'.annotation_short")
			Read from file... 'directory$'/'name$'.annotation_short
			
			#Duplicate and remove all boundaries
			Copy... 'name$'_annot
			select TextGrid 'name$'_annot
			annot_ntier = Get number of tiers
			for i from 1 to annot_ntier
				annot_ninterval = Get number of intervals... i
				n = annot_ninterval - 1
				for j from 1 to n
					Remove right boundary... i 1
				endfor
			endfor
			
			select TextGrid 'name$'
			ntiers = Get number of tiers
		else
			#[EXIT_A09] Annotation file 'name$' not found
			exit Error Code A09: Annotation file 'name$' not found
		endif
	elsif task = taskCreate
		select Sound 'name$'
		start_time = Get start time
		end_time = Get end time

		To TextGrid... "'list_func$'"
		ntiers = Get number of tiers
		
		Copy... 'name$'_annot
		select TextGrid 'name$'
			
		if copy_from$ != ""
			Read from file... 'directory$'/'name$''copy_from$'
			Rename... _boundary_
			nintervals = Get number of intervals... 'copy_tier'
			
			for j from 2 to nintervals
				select TextGrid _boundary_
				start = Get start point... 'copy_tier' j
				select TextGrid 'name$'
				Insert boundary... 1 start
				if copy_content > 0
					select TextGrid _boundary_
					labc$ = Get label of interval... 'copy_tier' j
					if labc$ <> ""
						select TextGrid 'name$'
						Set interval text... 1 j 'labc$'
					endif
				endif
			endfor
			
			select TextGrid _boundary_
			Remove
		endif
	elsif task = taskAdd
		if fileReadable ("'directory$'/'name$'.annotation_short")
			Read from file... 'directory$'/'name$'.annotation_short
			ntiers = Get number of tiers
			newtier_number = ntiers+1
			Insert interval tier... newtier_number 'function_name$'
			nintervals = Get number of intervals... 1
			
			for j from 2 to nintervals
				start = Get starting point... 1 j
				Insert boundary... newtier_number start
			endfor
			
			#Duplicate and remove all boundaries
			Copy... 'name$'_annot
			select TextGrid 'name$'_annot
			annot_ntier = Get number of tiers
			for i from 1 to annot_ntier
				annot_ninterval = Get number of intervals... i
				n = annot_ninterval - 1
				for j from 1 to n
					Remove right boundary... i 1
				endfor
			endfor
		else
			exit Annotation file 'name$' not found
		endif
	elsif task = taskRemove
		if tier_number > 0
			if fileReadable ("'directory$'/'name$'.annotation_short")
				Read from file... 'directory$'/'name$'.annotation_short
				Remove tier... tier_number
				ntiers = Get number of tiers
				
				#Duplicate and remove all boundaries
				Copy... 'name$'_annot
				select TextGrid 'name$'_annot
				annot_ntier = Get number of tiers
				for i from 1 to annot_ntier
					annot_ninterval = Get number of intervals... i
					n = annot_ninterval - 1
					for j from 1 to n
						Remove right boundary... i 1
					endfor
				endfor
			else
				exit Annotation file 'name$' not found
			endif
		endif
	endif
	
	if task <> taskRemove 
		if process_all_files_without_pause = 0
			select TextGrid 'name$'
			plus Sound 'name$'
			Edit
		 
		
			pulsefile$ = name$+".pulse"
			if fileReadable ("'directory$'/'pulsefile$'")
				Read from file... 'directory$'/'pulsefile$'
			else
				select Sound 'name$'
				To PointProcess (periodic, cc)... minf0 maxf0
				Rename... 'name$'
			endif
		
			if inspect_pulse
				plus Sound 'name$'
				Edit
			endif
		
			pause Click Continue when you finished editing
		else
			printline Processing 'name$'
			pulsefile$ = name$+".pulse"
			if fileReadable ("'directory$'/'pulsefile$'")
				Read from file... 'directory$'/'pulsefile$'
			else
				select Sound 'name$'
				To PointProcess (periodic, cc)... minf0 maxf0
				Rename... 'name$'
			endif
		endif
	else
		pulsefile$ = name$+".pulse"
		if fileReadable ("'directory$'/'pulsefile$'")
			Read from file... 'directory$'/'pulsefile$'
		else
			select Sound 'name$'
			To PointProcess (periodic, cc)... minf0 maxf0
			Rename... 'name$'
		endif
	endif
	
	call CreateFullCombination 'name$'
	
	replot = 1
	call Save 'directory$' 'name$'
	
	select Sound 'name$'
	Remove
endproc


###Clean all files generated for the target file
procedure deleteAuxFiles name$
	if fileReadable("'directory$'/'name$'.actutimenormf0")
		deleteFile("'directory$'/'name$'.actutimenormf0")
	endif
	if fileReadable("'directory$'/'name$'.annotation")
		deleteFile("'directory$'/'name$'.annotation")
	endif
	if fileReadable("'directory$'/'name$'.annotation_short")
		deleteFile("'directory$'/'name$'.annotation_short")
	endif
	if fileReadable("'directory$'/'name$'.f0")
		deleteFile("'directory$'/'name$'.f0")
	endif
	if fileReadable("'directory$'/'name$'.f0velocity")
		deleteFile("'directory$'/'name$'.f0velocity")
	endif
	if fileReadable("'directory$'/'name$'.PitchTier")
		deleteFile("'directory$'/'name$'.PitchTier")
	endif
	if fileReadable("'directory$'/'name$'.pulse")
		deleteFile("'directory$'/'name$'.pulse")
	endif
	if fileReadable("'directory$'/'name$'.rawf0")
		deleteFile("'directory$'/'name$'.rawf0")
	endif
	if fileReadable("'directory$'/'name$'.samplef0")
		deleteFile("'directory$'/'name$'.samplef0")
	endif
	if fileReadable("'directory$'/'name$'.semitonef0")
		deleteFile("'directory$'/'name$'.semitonef0")
	endif
	if fileReadable("'directory$'/'name$'.timenormf0")
		deleteFile("'directory$'/'name$'.timenormf0")
	endif
endproc

###Save the auxiliary files related to name$###
procedure CreateFullCombination name$
	#Propagate the smallest tier to other tiers
	select TextGrid 'name$'
	ntier_template = Get number of tiers
	if ntier_template > 1
		#Search for the smallest tier
		s_tier = 1
		s_n = Get number of intervals... 1
		for i from 2 to ntier_template
			x_n = Get number of intervals... i
			if x_n > s_n
				s_tier = i
				s_n = x_n
			endif
		endfor                  
		
		#Insert boundaries to other tier
		for i from 1 to ntier_template
			for j from 1 to s_n
				select TextGrid 'name$'
				t_end = Get end point... s_tier j
				select TextGrid 'name$'_annot
				if (j<s_n)
					Insert boundary... i t_end
				endif	
			endfor
		endfor
		
		#Copy labels from .annotation_short
		for i from 1 to ntier_template
			for j from 1 to s_n
				select TextGrid 'name$'
				t_start = Get starting point... s_tier j
				t_end = Get end point... s_tier j
				t_mid = (t_start + t_end)/2
				x_int = Get interval at time... i t_mid
				label$ = Get label of interval... i x_int
				select TextGrid 'name$'_annot
				Set interval text... i j 'label$'
			endfor
		endfor
		
		#Delete intervals that should be voiceless
		for j from 1 to s_n
			select TextGrid 'name$'
			t_start = Get starting point... s_tier j
			t_end = Get end point... s_tier j
			t_mid = (t_start + t_end)/2
			x_int = Get interval at time... s_tier t_mid
			label$ = Get label of interval... s_tier x_int
			if label$ = ""
				select TextGrid 'name$'_annot
				for i from 1 to ntier_template
					Set interval text... i j 'label$'
				endfor
			endif				
		endfor
	elsif ntier_template = 1
		#Search for the smallest tier                  
		s_tier = 1
		s_n = Get number of intervals... 1
		
		#Insert boundaries to other tier
		for j from 1 to s_n
			select TextGrid 'name$'
			t_end = Get end point... s_tier j
			select TextGrid 'name$'_annot
			if (j<s_n)
				Insert boundary... 1 t_end
			endif
		endfor
		
		#Copy labels from .annotation_short
		for j from 1 to s_n
			select TextGrid 'name$'
			t_start = Get starting point... s_tier j
			t_end = Get end point... s_tier j
			t_mid = (t_start + t_end)/2
			x_int = Get interval at time... 1 t_mid
			label$ = Get label of interval... 1 x_int
			select TextGrid 'name$'_annot
			Set interval text... 1 j 'label$'
		endfor
		
		#Delete intervals that should be voiceless
		for j from 1 to s_n
			select TextGrid 'name$'
			t_start = Get starting point... s_tier j
			t_end = Get end point... s_tier j
			t_mid = (t_start + t_end)/2
			x_int = Get interval at time... s_tier t_mid
			label$ = Get label of interval... s_tier x_int
			if label$ = ""
				select TextGrid 'name$'_annot
				Set interval text... 1 j 'label$'
			endif				
		endfor
	endif
endproc

procedure Save directory$ name$
	select PointProcess 'name$'
	npulses = Get number of points
	for n from 2 to npulses
		time1 = Get time from index... n-1
		time2 = Get time from index... n
		if time2 = time1
			Remove point... n
		endif
	endfor
	Write to short text file... 'directory$'/'name$'.pulse
	maxperiod = 1/minf0
	To PitchTier... maxperiod
	Down to TableOfReal... Hertz
	Write to headerless spreadsheet file... 'directory$'/'name$'.rawf0
	Remove
	select PitchTier 'name$'
	call Trimf0
	Write to short text file... 'directory$'/'name$'.PitchTier
	Down to TableOfReal... Hertz
	Write to headerless spreadsheet file... 'directory$'/'name$'.f0
	Remove

	# Create sampled F0 and velocity with a constant sampling rate
	call Sampling
	if found_interval
		call Differentiation
		call Smooth "velocity" "smooth_velocity" 0.05 * f0_sample_rate
		select TableOfReal samplef0
		Write to headerless spreadsheet file... 'directory$'/'name$'.samplef0
		Remove
		select TableOfReal velocity
		Write to headerless spreadsheet file... 'directory$'/'name$'.f0velocity
		plus PitchTier smoothf0
		Remove
	endif

	# Create sampled F0 with a constant number per interval
	if found_interval
		call Normalization
		
		call CheckF0
		
		select TableOfReal normf0
		Write to headerless spreadsheet file... 'directory$'/'name$'.timenormf0
		select TableOfReal normactuf0
		Write to headerless spreadsheet file... 'directory$'/'name$'.actutimenormf0
		plus TableOfReal normf0
		plus TableOfReal normactuf0
		Remove
		
	;	call Means
	;	select PitchTier semitonef0
	;	Write to short text file... 'directory$'/'name$'.PitchTier_semitone
		select PitchTier smooth_velocity
		Remove

	;	select TableOfReal means
	;	Write to headerless spreadsheet file... 'directory$'/'name$'.means
	;	plus Intensity 'name$'
	;	Remove
		select PitchTier 'name$'
		plus PitchTier velocity
		Remove
	endif
	
	select TextGrid 'name$'
	Write to short text file... 'directory$'/'name$'.annotation_short

	select TextGrid 'name$'_annot
	Write to short text file... 'directory$'/'name$'.annotation
	
	select TextGrid 'name$'_annot
	Down to Table... no 4 yes no
	Save as tab-separated file... 'directory$'/'name$'.intervals
	
#	nfunc = ntiers
#	Create Table with column names... func 'nfunc' FUNCTION

#	for k from 1 to nfunc
#		select TextGrid 'name$'
#		;j = k+1
#		fname$ = Get tier name... 'k'
#		select Table func
#		Set string value... 'k' FUNCTION 'fname$'
#	endfor
	
#	select Table func
#	if task <> taskAdd
#		Save as tab-separated file... 'directory$'/function.txt
#	endif
#	Remove
	
	select TextGrid 'name$'
	plus TextGrid 'name$'_annot
	plus Table 'name$'_annot
	plus PointProcess 'name$'
	if found_interval
		plus PitchTier semitonef0
	;	plus PitchTier fittedf0
	;	plus Manipulation 'name$'
	;else
	;	plus PitchTier 'name$'
	;	plus PitchTier samplef0
	;	plus TableOfReal samplef0
	endif
	Remove
endproc


procedure Trimf0
	maxbump = 0.01
	maxedge = 0.0
	maxgap = 0.033
	n = Get number of points
	
	first = Get value at index... 1
	second = Get value at index... 2
	penult = Get value at index... n-1
	last = Get value at index... n
	tfirst = Get time from index... 1
	tlast = Get time from index... n
	for k from 1 to 3
		call Trim
	endfor
	Remove point... 1
	Add point... tfirst second + (first-second) / 1000
	Remove point... n
	Add point... tlast penult + (last-penult) / 1000
endproc

procedure Trim
	for i from 2 to n-1
		tleft = Get time from index... i-1
		tmid = Get time from index... i
		tright = Get time from index... i+1
		gap1 = tmid - tleft
		gap2 = tright - tmid
		left = Get value at index... i-1
		mid = Get value at index... i
		right = Get value at index... i+1
		diff1 = mid - left
		diff2 = mid - right
		if diff1 > maxbump and diff2 > maxedge and gap1 < maxgap and gap2 < maxgap
		... or diff2 > maxbump and diff1 > maxedge and gap1 < maxgap and gap2 < maxgap
			Remove point... i
			Add point... tmid left+(tmid-tleft)/(tright-tleft)*(right-left)
		endif
		if diff1 > maxbump and gap2 >= maxgap
			Remove point... i
			Add point... tmid left + maxbump
		endif
		if diff2 > maxbump and gap1 >= maxgap
			Remove point... i
			Add point... tmid right + maxbump
		endif

		diff1 = left - mid
		diff2 = right - mid
		if diff1 > maxbump and diff2 > maxedge and gap1 < maxgap and gap2 < maxgap
		... or diff2 > maxbump and diff1 > maxedge and gap1 < maxgap and gap2 < maxgap
			Remove point... i
			Add point... tmid left+(tmid-tleft)/(tright-tleft)*(right-left)
		endif
		if diff1 > maxbump and gap2 >= maxgap
			Remove point... i
			Add point... tmid left - maxbump
		endif
		if diff2 > maxbump and gap1 >= maxgap
			Remove point... i
			Add point... tmid right - maxbump
		endif
	endfor
endproc

procedure Normalization
	select PitchTier 'name$'
	lasttime = Get finishing time
	
	# create empty PitchTiers named "normactutimef0" and "normactuf0"
	Create PitchTier... normactutimef0 0 lasttime
	Create TableOfReal... normf0 1 2
	Set column label (index)... 1 NormalizedTime
	Set column label (index)... 2 F0
	Create TableOfReal... normactuf0 1 2
	Set column label (index)... 1 ActualTime
	Set column label (index)... 2 F0
	
	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	interval = 0
	found_interval = 0
	nrows = 0
	# for each interval
	for m from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 m
		if not label$ = ""
			start = Get starting point... 1 m
			end = Get end point... 1 m
			select PitchTier 'name$'
			index1 = Get high index from time... start
			index2 = Get low index from time... end
			if found_interval = 0
				found_interval = 1
				select PitchTier 'name$'
				firstime = start
			endif
			call Normalize
			interval = interval + 1
		endif
	endfor
	select TableOfReal normf0
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
	select TableOfReal normactuf0
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
	select PitchTier normactutimef0
	Remove
endproc

procedure Normalize
	duration = end - start

	for i from index1 to index2
		select PitchTier 'name$'
		time = Get time from index... i
		f0 = Get value at index... i
		select PitchTier normactutimef0
		Add point... (time-start)/duration+interval f0
	endfor
	
	for x from 1 to npoints
		normtime = x / npoints
		select PitchTier normactutimef0
		f0 = Get value at time... normtime+interval
		select TableOfReal normf0
		nrows = Get number of rows
		Set value... nrows 1 x+interval*npoints
		Set value... nrows 2 f0
		Set row label (index)... nrows 'label$'
		Insert row (index)... nrows + 1
		select TableOfReal normactuf0
		nrows = Get number of rows
		Set value... nrows 1 normtime*duration+start
		if set_initial_time_for_normf0_to_0
			Set value... nrows 1 normtime*duration+start-firstime
		endif
		Set value... nrows 2 f0
		Set row label (index)... nrows 'label$'
		Insert row (index)... nrows + 1
	endfor
endproc

# Sampling calculates the number of syllable from the sampling rate and samples F0 of the specified intervals
procedure Sampling
	# create an empty TableOfReal named "samplef0" with only the header
	select PitchTier 'name$'
	Create TableOfReal... samplef0 1 2
	Set column label (index)... 1 SampleTime
	Set column label (index)... 2 F0

	# extract the number of intervals from TextGrid annotation and initiate related variables
	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	nlabels = 0
	found_interval = 0
	nrows = 0
	
	# create an empty PitchTier named "samplef0" starting from sampleStart to sampleEnd
	sampleStart = Get starting point... 1 1
	sampleEnd = Get end point... 1 nintervals
	Create PitchTier... samplef0 sampleStart sampleEnd
	
	# for each interval
	for m from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 m
		if not label$ = ""
			nlabels += 1
			# get the start and end time of the interval, calculate the duration, and set the found_interval flag to 1
			start = Get starting point... 1 m
			end = Get end point... 1 m
			duration = end - start
			if found_interval = 0
				found_interval = 1
			endif

			# calculate the number of samples for the inverval
			select PitchTier 'name$'
			nsamples = duration * f0_sample_rate + 1
			
			# for each sampled time
			for x from 0 to nsamples - 1
				# get sample_time and f0 at that time
				select PitchTier 'name$'
				sample_time = start + x/f0_sample_rate
				f0 = Get value at time... sample_time
				
				# insert sample_time and f0 to the PitchTier "samplef0" and to the TableOfReal "samplef0"
				select PitchTier samplef0
				Add point... sample_time f0
				select TableOfReal samplef0
				nrows = Get number of rows
				Set value... nrows 1 sample_time
				Set value... nrows 2 f0
				Set row label (index)... nrows 'label$'
				Insert row (index)... nrows + 1
			endfor
		endif
	endfor
	select TableOfReal samplef0
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
endproc


procedure Differentiation
	select PitchTier samplef0
	Rename... semitonef0
	Formula... 12 * ln (self) / ln(2); semitone
	Down to TableOfReal... Hertz
	Set column label (index)... 1 time
	Set column label (index)... 2 f0 (st)

	Write to headerless spreadsheet file... 'directory$'/'name$'.semitonef0
	Remove
	call Smooth "semitonef0" "smoothf0" 0.05 * f0_sample_rate
	Create PitchTier... velocity sampleStart sampleEnd
	Create TableOfReal... velocity 1 2
	Set column label (index)... 1 SampleTime
	Set column label (index)... 2 Velocity(semitone/s)
	
	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	nrows = 0
	for m from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 m
		if not label$ = ""
			start = Get starting point... 1 m
			end = Get end point... 1 m
			select PitchTier smoothf0
			index_first = Get high index from time... start
			index_last = Get low index from time... end
			for x from index_first to index_last - 1
				if x = index_first or x = index_last - 1
					x2 = x + 1
				else
					x2 = x + 2
				endif
				select PitchTier smoothf0
				f01 = Get value at index... x
				f02 = Get value at index... x2
				sampletime1 = Get time from index... x
				sampletime2 = Get time from index... x2
				velocity = (f02 - f01) / (sampletime2 - sampletime1)
				velocity_time = 0.5 * (sampletime1 + sampletime2)
				select PitchTier velocity
				Add point... velocity_time velocity
				select TableOfReal velocity
				nrows = Get number of rows
				Set value... nrows 1 velocity_time
				Set value... nrows 2 velocity
				Set row label (index)... nrows 'label$'
				Insert row (index)... nrows + 1
			endfor
		endif
	endfor
	select TableOfReal velocity
	if nrows > 1
		Remove row (index)... nrows + 1
	endif
endproc

procedure Means
	select Sound 'name$'
	To Intensity... 100 0 yes

	select TextGrid 'name$'
	nintervals = Get number of intervals... 1
	Create TableOfReal... means 1 16
	Set column label (index)... 1 maxf0
	Set column label (index)... 2 minf0
	Set column label (index)... 3 excursion_size
	Set column label (index)... 4 meanf0
	Set column label (index)... 5 finalf0
	Set column label (index)... 6 mean_intensity
	Set column label (index)... 7 duration
	Set column label (index)... 8 max_velocity
	Set column label (index)... 9 final_velocity
	Set column label (index)... 10 initialf0
	Set column label (index)... 11 target_slope
	Set column label (index)... 12 target_height
	Set column label (index)... 13 strength
	Set column label (index)... 14 duration
	Set column label (index)... 15 rmse
	Set column label (index)... 16 correlation

	interval = 0
	for mm from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 mm
		if not label$ = ""
			interval = interval + 1
			start = Get starting point... 1 mm
			end = Get end point... 1 mm
			duration = 1000 * (end - start)
			start1 = start + perturbation_length/1000
			select TableOfReal means
			Set row label (index)... interval 'label$'
			select PitchTier 'name$'
			meanf0 = Get mean (points)... start1 end
			early_end = end + final_offset
			finalf0 = Get value at time... early_end
			To Pitch... 0.02 30 600
			max_f0 = Get maximum... start1 end Hertz Parabolic
			min_f0 = Get minimum... start1 end Hertz Parabolic
			excursionsize = hertzToSemitones(max_f0) - hertzToSemitones(min_f0)
			Remove
			select Intensity 'name$'
			intensity = Get mean... start end energy
			select PitchTier smooth_velocity
			final_velocity'interval' = Get value at time... early_end
			
			mid = 0.5 * (start + end)
			select PitchTier smooth_velocity
			index_first = Get high index from time... mid
			index_last = Get low index from time... end
			maxvelocity = 0
			for x from index_first to index_last
				v = Get value at index... x
				if abs(v) > abs(maxvelocity)
					maxvelocity = v
				endif
			endfor

			select TableOfReal means
			Set value... interval 1 max_f0
			Set value... interval 2 min_f0
			Set value... interval 3 excursionsize
			Set value... interval 4 meanf0
			Set value... interval 5 finalf0
			Set value... interval 6 intensity
			Set value... interval 7 duration
			Set value... interval 8 maxvelocity
			Set value... interval 9 final_velocity'interval'
			Insert row (index)... interval + 1
		endif
	endfor

	if found_interval
		call Target
		select Sound 'name$'
		To Manipulation... 0.01 75 600
		select PitchTier fittedf0
		Formula... exp(self*ln(2)/12)
		plus Manipulation 'name$'
		Replace pitch tier
		select TableOfReal fittedf0
		Write to headerless spreadsheet file... 'directory$'/'name$'.qTAf0
#		plus PitchTier fittedf0
		Remove
		select TableOfReal fittedvelocity
		Write to headerless spreadsheet file... 'directory$'/'name$'.qTAvelocity
		plus PitchTier fittedvelocity
		plus TableOfReal trainingf0
;		plus TableOfReal semitonef0
		Remove

		interval = 0
		for mm from 1 to nintervals
			select TextGrid 'name$'
			label$ = Get label of interval... 1 mm
			if not label$ = ""
				interval = interval + 1
				select TableOfReal means
				Set value... interval 10 xInitial
				Set value... interval 11 mBest'interval'
				Set value... interval 12 bBest'interval'
				Set value... interval 13 lambdaBest'interval'
				Set value... interval 14 duration'interval'
				Set value... interval 15 rmse'interval'
				Set value... interval 16 corr'interval'
			endif
		endfor
	endif
	select TableOfReal means
	nrows = Get number of rows
	if nrows > 1
		Remove row (index)... nrows
	endif
endproc

procedure Smooth curve_in$ curve_out$ width
	Create PitchTier... 'curve_out$' sampleStart sampleEnd
	
	for j from 1 to width							; make a triangular window of size = width 
		if j < width / 2 + 0.5
			weight'j' = j
		else 
			weight'j' = width - j + 1
		endif
	endfor

	select PitchTier 'curve_in$'
	smooth_end = Get number of points

	for i from 1 to width / 2						; smooth initial points: 0 to width/2 - 1 
	    n = 0.0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... i
	    for j from 1 to width/2 + i					; window size = width/2 to width - 1
	    	j = 'j:0'
	    	rawsample = Get value at index... j
	    	index = width / 2 + j - i
	    	index = 'index:0'
			smoothsample += weight'index' * rawsample
			n += weight'index'
	    endfor
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
	endfor
	
	for i from width/2 to smooth_end - width/2				; smooth from width/2 to end-width/2
		n = 0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... i
		for j from 1 to width
	    	rawsample = Get value at index... i-width/2+j
			smoothsample += weight'j' * rawsample
			n += weight'j'
		endfor
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
	endfor
	
	i = width/2
	while i > 0										; smooth final points: end - width/2 to end
		n = 0.0
		smoothsample = 0.0
		select PitchTier 'curve_in$'
		sample_time = Get time from index... smooth_end-i
		j = width/2 + i
		j = 'j:0'
		while j > 1									; window size = width - 1 to width/2 
	    	rawsample = Get value at index... smooth_end-j
	    	index = width/2+i-j + 1
	    	index = 'index:0'
			smoothsample += weight'index' * rawsample
			n += weight'index'
			j -= 1
		endwhile
		smoothsample /= n
		if smoothsample != undefined
			select PitchTier 'curve_out$'
			Add point... sample_time smoothsample
		endif
		i -= 1
	endwhile
endproc

procedure CheckF0
	flag = 1
	select TableOfReal normf0
	nrows = Get number of rows
	for i from 1 to nrows
		fx = Get value... i 2
		if fx = undefined
			flag = 0
		endif
	endfor
	
	select TableOfReal normactuf0
	nrows = Get number of rows
	for i from 1 to nrows
		fx = Get value... i 2
		if fx = undefined
			flag = 0
		endif
	endfor
		
	select Table f0check
	Set string value... current_file Name 'name$'
	Set numeric value... current_file Flag flag
endproc

procedure ReportError
	cflag = 1
	select Table f0check
	nrows = Get number of rows
	for i from 1 to nrows
		select Table f0check
		flag = Get value... i Flag
		name$ = Get value... i Name
		if flag = 0
			printline Error: undefined value at file: 'name$'
			cflag = 0
		endif
	endfor
endproc
